@@ -669,45 +669,43 @@ sub extract_bracketed_text($)
   my $str = "";
 
   # Move to the first "(":
-  while ( scalar(@tokens) )
-    {
-      my $t = $self->extract_token();
-      last if ($t eq "(");
-    }
+  while (scalar(@tokens))
+  {
+    my $t = $self->extract_token();
+    last if ($t eq "(");
+  }
 
-  # TODO: Don't count "(" and ")" within double quotes.
-  # There may be .hg files with unpaired quotes that generate correct
-  # .h and .cc files. Don't want to break such code yet.
-  # See also TODO in string_split_commas().
+  my $filename = $$self{filename};
+  my $line_num = $$self{line_num};
 
   # Concatenate until the corresponding ")":
-  while ( scalar(@tokens) )
+  while (scalar(@tokens) and $filename eq $$self{filename})
+  {
+    my $t = $self->extract_token();
+    $in_quotes = !$in_quotes if ($t eq '"');
+    # Don't count "(" and ")" within double quotes.
+    if (!$in_quotes)
     {
-      my $t = $self->extract_token();
-      $in_quotes = !$in_quotes if ($t eq '"');
       $level++ if ($t eq "(");
       $level-- if ($t eq ")");
-
-      if (!$level)
-      {
-        $self->error("End of gmmproc directive within a quoted string.\n") if $in_quotes;
-        return $str;
-      }
-      $str .= $t;
+      return $str if (!$level); # Found matching ")"
     }
+    $str .= $t;
+  }
 
-  return "";
+  # No matching ")" found.
+  my $quote_text = $in_quotes ? " in a quoted string" : "";
+  die "$filename:$line_num: *** End of file$quote_text in a gmmproc macro. ***\n";
 }
 
 
 ########################################
 ###  breaks up a string by commas (smart)
-# @strings string_split_commas($string [, $ignore_quotes])
-sub string_split_commas($;$)
+# @strings string_split_commas($string)
+sub string_split_commas($)
 {
-  my ($in, $ignore_quotes) = @_;
+  my ($in) = @_;
 
-  $ignore_quotes = 2 unless defined $ignore_quotes;
   my @out;
   my $level = 0;
   my $in_quotes = 0;
@@ -720,10 +718,7 @@ sub string_split_commas($;$)
 
     next if ($t eq "");
 
-    # TODO: Delete the test for scalar(@out) >= $ignore_quotes when we can stop accepting
-    # .hg files with unpaired quotes, such as _WRAP_PROPERTY("text_column, int).
-    # See also TODO in extract_bracketed_text().
-    $in_quotes = !$in_quotes if ($t eq '"' and scalar(@out) >= $ignore_quotes);
+    $in_quotes = !$in_quotes if ($t eq '"');
     if (!$in_quotes)
     {
       $level++ if ($t eq "(" or $t eq "<" or $t eq "{");
@@ -997,8 +992,8 @@ sub on_wrap_method($)
     }
     # The "slot_callback" option tells gmmproc the name of the
     # callback function that should be passed to the C function if the
-    # method has a slot.
-    elsif($argRef =~ /^slot_callback\s+(\w+)/)
+    # method has a slot. The name can contain a namespace prefix.
+    elsif($argRef =~ /^slot_callback\s+([:\w]+)/)
     {
       $$objCppfunc{slot_callback} = $1;
     }
@@ -1035,7 +1030,7 @@ sub on_wrap_method_docs_only($)
   my $line_num = $$self{line_num};
 
   my $str = $self->extract_bracketed_text();
-  my @args = string_split_commas($str, 1);
+  my @args = string_split_commas($str);
 
   my $entity_type = "method";
 
@@ -1446,6 +1441,7 @@ sub on_wrap_any_enum($$)
   my @subst_in  = [];
   my @subst_out = [];
   my $gtype_func = "";
+  my $conv_to_int = "";
   my $argDeprecated = "";
   my $deprecation_docs = "";
   my $newin = "";
@@ -1467,12 +1463,16 @@ sub on_wrap_any_enum($$)
 
     if ($arg eq "NO_GTYPE")
     {
-      $gtype_func = "NO_GTYPE";
+      $gtype_func = $arg;
     }
     elsif ($arg =~ /^gtype_func\s+(\w+)/)
     {
       $gtype_func = $1;
     }
+    elsif (!$is_gerror and $arg eq "CONV_TO_INT")
+    {
+      $conv_to_int = $arg;
+    }
     elsif ($arg =~ /^s#([^#]+)#([^#]*)#$/)
     {
       push(@subst_in,  $1);
@@ -1497,9 +1497,31 @@ sub on_wrap_any_enum($$)
     }
   }
   return ($cpp_type, $c_type, $domain, \@subst_in, \@subst_out, $gtype_func,
-    $argDeprecated, $deprecation_docs, $newin, $decl_prefix);
+    $conv_to_int, $argDeprecated, $deprecation_docs, $newin, $decl_prefix);
 }
 
+# void on_wrap_enum()
+# _WRAP_ENUM(cpp_type, c_type [,NO_GTYPE] [,gtype_func funcname] [,CONV_TO_INT] [,s#regexpr#subst#]*)
+# Optional arguments:
+# NO_GTYPE Don't generate code for a specialization of the template
+#          Glib::Value_Enum or Glib::Value_Flags.
+#          Necessary, if the C type enum is not registered as a GType.
+# gtype_func funcname Call funcname() in the generated Glib::Value<>::value_type().
+#                     Necessary, if the M4 macro _GET_TYPE_FUNC() can't generate
+#                     the correct function name from the C type name of the enum.
+# CONV_TO_INT "Convertible to int" Generate a plain enum (not an enum class)
+#             within a class. Such an enum is scoped like an enum class,
+#             and it can be implicitly converted to int like a plain enum.
+# s#regexpr#subst# Zero or more substitutions in names of enum constants, e.g. s#^DATE_##.
+#
+# _WRAP_ENUM can be located either in a class or outside all classes.
+# When located in a class, and Value specialization shall be generated or it's
+# a Flags type (i.e. bitwise operators shall be generated), then the following
+# requirements must be fulfilled:
+# 1. _WRAP_ENUM must be located in the public part of the class.
+# 2. The class must contain a class macro (_CLASS_GENERIC, _CLASS_GOBJECT,
+#    _CLASS_GTKOBJECT, etc.) before _WRAP_ENUM.
+#
 sub on_wrap_enum($)
 {
   my ($self) = @_;
@@ -1509,13 +1531,13 @@ sub on_wrap_enum($)
   my $comment = $self->extract_preceding_documentation();
 
   # get the arguments
-  my ($cpp_type, $c_type, undef, $ref_subst_in, $ref_subst_out, $gtype_func,
+  my ($cpp_type, $c_type, undef, $ref_subst_in, $ref_subst_out, $gtype_func, $conv_to_int,
     $argDeprecated, $deprecation_docs, $newin, $decl_prefix) = $self->on_wrap_any_enum(0);
 
   $$self{objOutputter}->output_wrap_enum(
     $$self{filename}, $$self{line_num}, $cpp_type, $c_type,
-    $comment, $ref_subst_in, $ref_subst_out, $gtype_func,
-    $argDeprecated, $deprecation_docs, $newin, $decl_prefix);
+    $comment, $ref_subst_in, $ref_subst_out, $gtype_func, $conv_to_int,
+    $$self{in_class}, $argDeprecated, $deprecation_docs, $newin, $decl_prefix);
 }
 
 sub on_wrap_enum_docs_only($)
@@ -1527,7 +1549,7 @@ sub on_wrap_enum_docs_only($)
   my $comment = $self->extract_preceding_documentation();
 
   # get the arguments
-  my ($cpp_type, $c_type, undef, $ref_subst_in, $ref_subst_out, undef,
+  my ($cpp_type, $c_type, undef, $ref_subst_in, $ref_subst_out, undef, $conv_to_int,
     $argDeprecated, $deprecation_docs, $newin) = $self->on_wrap_any_enum(0);
 
   # Get the module name so the enum docs can be included in the module's
@@ -1536,7 +1558,8 @@ sub on_wrap_enum_docs_only($)
 
   $$self{objOutputter}->output_wrap_enum_docs_only(
     $$self{filename}, $$self{line_num}, $module_canonical, $cpp_type, $c_type,
-    $comment, $ref_subst_in, $ref_subst_out, $deprecation_docs, $newin);
+    $comment, $ref_subst_in, $ref_subst_out, $conv_to_int, $$self{in_class},
+    $deprecation_docs, $newin);
 }
 
 sub on_wrap_gerror($)
@@ -1551,7 +1574,7 @@ sub on_wrap_gerror($)
   my $class_docs = $self->extract_preceding_documentation();
 
   # get the arguments
-  my ($cpp_type, $c_type, $domain, $ref_subst_in, $ref_subst_out, $gtype_func,
+  my ($cpp_type, $c_type, $domain, $ref_subst_in, $ref_subst_out, $gtype_func, undef,
     $argDeprecated, $deprecation_docs, $newin, $decl_prefix) = $self->on_wrap_any_enum(1);
 
   $$self{objOutputter}->output_wrap_gerror(
@@ -1588,6 +1611,7 @@ sub on_wrap_any_property($)
   my $argDeprecated = "";
   my $deprecation_docs = "";
   my $newin = "";
+  my $bNoTypeCheck = 0;
   while($#args >= 2) # If the optional arguments are there.
   {
     my $argRef = string_trim(pop @args);
@@ -1605,10 +1629,14 @@ sub on_wrap_any_property($)
     {
       $newin = string_unquote(string_trim($1));
     }
+    elsif($argRef eq "no_type_check")
+    {
+      $bNoTypeCheck = 1;
+    }
   }
 
   return ($filename, $line_num, $argPropertyName, $argCppType,
-          $argDeprecated, $deprecation_docs, $newin);
+          $argDeprecated, $deprecation_docs, $newin, $bNoTypeCheck);
 }
 
 sub on_wrap_property($)
@@ -1619,10 +1647,11 @@ sub on_wrap_property($)
   return unless ($self->check_for_eof());
 
   my ($filename, $line_num, $argPropertyName, $argCppType, $argDeprecated,
-      $deprecation_docs, $newin) = $self->on_wrap_any_property();
+      $deprecation_docs, $newin, $bNoTypeCheck) = $self->on_wrap_any_property();
 
   $objOutputter->output_wrap_property($filename, $line_num, $argPropertyName,
-    $argCppType, $$self{c_class}, $$self{deprecated}, $argDeprecated, $deprecation_docs, $newin);
+    $argCppType, $$self{c_class}, $$self{deprecated}, $argDeprecated, $deprecation_docs,
+    $newin, $bNoTypeCheck);
 }
 
 sub on_wrap_child_property($)
@@ -1633,10 +1662,11 @@ sub on_wrap_child_property($)
   return unless ($self->check_for_eof());
 
   my ($filename, $line_num, $argPropertyName, $argCppType, $argDeprecated,
-      $deprecation_docs, $newin) = $self->on_wrap_any_property();
+      $deprecation_docs, $newin, $bNoTypeCheck) = $self->on_wrap_any_property();
 
   $objOutputter->output_wrap_child_property($filename, $line_num, $argPropertyName,
-    $argCppType, $$self{c_class}, $$self{deprecated}, $argDeprecated, $deprecation_docs, $newin);
+    $argCppType, $$self{c_class}, $$self{deprecated}, $argDeprecated, $deprecation_docs,
+    $newin, $bNoTypeCheck);
 }
 
 sub output_wrap_check($$$$$$)
